/*
 * Copy me if you can.
 * by 20h
 */

#include <u.h>
#include <libc.h>
#include <thread.h>
#include <fcall.h>
#include <9p.h>
#include "usb.h"

#define STACKSIZE	16 * 1024
#define BUFSIZE		1024

#define VEND_WR		0x01
#define VEND_WR_T	RH2D | Rclass | Rdevice
#define VEND_RD		0x01
#define	VEND_RD_T	RD2H | Rvendor | Rdevice
#define SET_LINE	0x20
#define	SET_LINE_T	RH2D | Rclass | Rinterface
#define GET_LINE	0x21
#define GET_LINE_T	RD2H | Rclass | Rinterface
#define SET_CTRL	0x22
#define SET_CTRL_T	RH2D | Rclass | Rinterface
#define BRK			0x23
#define	BRK_T		RH2D | Rclass | Rinterface

#define PATH(type, n)		((type)|((n)<<8))
#define TYPE(path)			((int)(path) & 0xFF)
#define NUM(path)			((uint)(path)>>8)

enum {
	Qroot,
	Quia,
	Quiactl,
	Quiastatus,
};

typedef struct Tab Tab;
struct Tab
{
	char *name;
	ulong mode;
};
Tab tab[] =
{
	"/",		DMDIR|0555,
	"uia",		0666,
	"uiactl",	0666,
	"uiastatus",0444,
};

char ident[]		= "Enabled 0x0000ff 0x067b 0x2303";
char *cfmt			= "/dev/usb%d/%d/%s";
char *instr			= "ep 3 bulk r 32 64";
char *outstr		= "ep 2 bulk w 32 64";
int maxqueue		= 256;

Req *rq;
Req **erq;
char termios[7];
char ctrl;
int debug;
int statfd;
int infd;
int outfd;
long time0;
QLock relock;

void
queuereq(Req *r)
{
	if(rq == nil)
		erq = &rq;
	*erq = r;
	r->aux = nil;
	erq = (Req **)&r->aux;
}

void
fsreadproc(void *)
{
	char buf[256];
	Req *r;
	int n;

	n = 0;

	for(;;) {
		if(n == 0) {
			n = read(infd, buf, sizeof(buf));
			if(n < 0)
				sysfatal("read: %r");
		}
		if(rq != nil) {
			qlock(&relock);
			r = rq;
			rq = r->aux;
			qunlock(&relock);

			if(r->ifcall.count < n) {
				memmove(r->ofcall.data, buf, r->ifcall.count);
				r->ofcall.count = r->ifcall.count;
				n -= r->ifcall.count;
				memmove(buf, buf + r->ifcall.count, n);
			} else {
				memmove(r->ofcall.data, buf, n);
				r->ofcall.count = n;
				n = 0;
			}
			respond(r, nil);
		} else
			sleep(1);
	}
}

int
unpack32(char *src)
{
	int ret;

	ret = (src[3] << 24)	& 0xFF000000;
	ret |= (src[2] << 16) 	& 0x00FF0000;
	ret |= (src[1] << 8) 	& 0x0000FF00;
	ret |= src[0] 			& 0x000000FF;

	return ret;
}

void
statuscmd(int type, int req, int value, int index, char *data, int count)
{
	char *wp;

	wp = emalloc9p(8 + count);

	wp[0] = type;
	wp[1] = req;
	PUT2(wp + 2, value);
	PUT2(wp + 4, index);
	PUT2(wp + 6, count);
	if(data != nil)
		memmove(wp + 8, data, count);
	if(write(statfd, wp, 8 + count) != 8 + count)
		sysfatal("statuscmd: %r");
	free(wp);
}

void
statusread(char *buf, int count)
{
	for(;;) {
		count = read(statfd, buf, count);
		if(count >= 0)
			break;
	}
	if(count < 0)
		sysfatal("statusread: %r");
}

void
setbaud(char *baudr)
{
	int baud;

	baud = atoi(baudr);
	if(baud) {
		termios[0] = baud & 0xFF;
		termios[1] = (baud >> 8) & 0xFF;
		termios[2] = (baud >> 16) & 0xFF;
		termios[3] = (baud >> 24) & 0xFF;
	}
}

void
setstopbits(char *stopbits)
{
	termios[4] = 0;
	if(strcmp(stopbits, "1.5") == 0)
		termios[4] = 1;
	if(strcmp(stopbits, "2") == 0)
		termios[4] = 2;
}

void
setparity(char *parity)
{
	termios[5] = 0;
	switch(*parity) {
	case 'o':
		termios[5] = 1;
		break;
	case 'e':
		termios[5] = 2;
		break;
	case 'm':
		termios[5] = 3;
		break;
	case 's':
		termios[5] = 4;
		break;
	default:
		break;
	};
}

void
setbpb(char *bpb)
{
	termios[6] = 8;
	if(strcmp(bpb, "5") == 0)
		termios[6] = 5;
	if(strcmp(bpb, "6") == 0)
		termios[6] = 6;
	if(strcmp(bpb, "7") == 0)
		termios[6] = 7;
}

void
settermios(void)
{
	statuscmd(SET_LINE_T, SET_LINE, 0, 0, termios, 7);
}

void
gettermios(void)
{
	statuscmd(GET_LINE_T, GET_LINE, 0, 0, nil, 0);
	statusread(termios, 7);
}

void
setctrl(void)
{
	statuscmd(SET_CTRL_T, SET_CTRL, ctrl, 0, nil, 0);
}

void
setdtr(char *dtrs)
{
	int dtr;

	dtr = atoi(dtrs);
	if(dtr)
		ctrl |= 0x01;
	else
		ctrl &= ~0x01;

	setctrl();
}

void
setrts(char *rtss)
{
	int rts;

	rts = atoi(rtss);
	if(rts)
		ctrl |= 0x02;
	else
		ctrl &= ~0x02;

	setctrl();
}

void
setbrk(char *brks)
{
	vlong tim, brk;

	brk = atoll(brks);
	if(brk < 0)
		brk = 200;
	tim = nsec();
	statuscmd(BRK_T, BRK, 0xFFFF, 0, nil, 0);
	while((tim - nsec()) / 1000000 < brk);
	statuscmd(BRK_T, BRK, 0x0000, 0, nil, 0);
}

void
initpl2303(void)
{
/*
 * From the Prolific PL2303 Linux driver.
 * These VEND_* calls do not seem to be needed, because the device
 * termio changes without them and setups the line correctly.
 * There are also unhandled cases in the original driver, that
 * are not handled here.
 */
/*
	statuscmd(VEND_RD_T, VEND_RD, 0x8484, 0, nil, 0);
	statuscmd(VEND_WR_T, VEND_WR, 0x0404, 0, nil, 0);
	statuscmd(VEND_RD_T, VEND_RD, 0x8484, 0, nil, 0);
	statuscmd(VEND_RD_T, VEND_RD, 0x8383, 0, nil, 0);
	statuscmd(VEND_RD_T, VEND_RD, 0x8484, 0, nil, 0);
	statuscmd(VEND_WR_T, VEND_WR, 0x0404, 0, nil, 1);
	statuscmd(VEND_RD_T, VEND_RD, 0x8484, 0, nil, 0);
	statuscmd(VEND_RD_T, VEND_RD, 0x8383, 0, nil, 0);
	statuscmd(VEND_WR_T, VEND_WR, 0, 1, nil, 0);
	statuscmd(VEND_RD_T, VEND_RD, 1, 0, nil, 0);
*/

	gettermios();
	setbaud("9600");
	setstopbits("1");
	setparity("n");
	setbpb("8");
	settermios();

	ctrl = 0;
	setctrl();
}

static void
fillstat(Dir *d, uvlong path)
{
	Tab *t;

	memset(d, 0, sizeof(*d));
	d->uid = estrdup9p("uia");
	d->gid = estrdup9p("uia");
	d->qid.path = path;
	d->atime = d->mtime = time0;
	t = &tab[TYPE(path)];
	d->name = estrdup9p(t->name);
	d->qid.type = t->mode >> 24;
	d->mode = t->mode;
}

static void
fsattach(Req *r)
{
	if(r->ifcall.aname && r->ifcall.aname[0]) {
		respond(r, "invalid attach specifier");
		return;
	}
	r->fid->qid.path = PATH(Qroot, 0);
	r->fid->qid.type = QTDIR;
	r->fid->qid.vers = 0;
	r->ofcall.qid = r->fid->qid;
	respond(r, nil);
}

static void
fsstat(Req *r)
{
	fillstat(&r->d, r->fid->qid.path);
	respond(r, nil);
}

static char *
fswalk1(Fid *fid, char *name, Qid *qid)
{
	int i;
	ulong path;

	path = fid->qid.path;
	if(!(fid->qid.type & QTDIR))
		return "walk in non-directory";

	if(strcmp(name, "..") == 0) {
		switch(TYPE(path)) {
		case Qroot:
			return nil;
		default:
			return "bug in fswalk1";
		}
	}

	i = TYPE(path) + 1;
	for(; i < nelem(tab); i++) {
		if(strcmp(name, tab[i].name) == 0) {
			qid->path = PATH(i, NUM(path));
			qid->type = tab[i].mode >> 24;
			return nil;
		}
		if(tab[i].mode & DMDIR)
			break;
	}

	return "directory entry not found";
}

static int
rootgen(int i, Dir *d, void *)
{
	i += Qroot + 1;
	if(i <= Quiastatus) {
		fillstat(d, i);
		return 0;
	}
	return -1;
}

static void
uiaread(void *r)
{
	qlock(&relock);
	queuereq((Req *)r);
	qunlock(&relock);
}

static void
uiawrite(void *re)
{
	Req *r;
	int n;

	r = re;
	n = write(outfd, r->ifcall.data, r->ifcall.count);
	if(n < 0) {
		respond(r, "write");
		return;
	}
	r->ofcall.count = n;

	respond(r, nil);
}

static void
uiactlread(Req *r)
{
	respond(r, nil);
}

static void
uiactlwrite(Req *r)
{
	char *cur, *lastp;

	if(r->ifcall.count < 1) {
		respond(r, "bad command");
		return;
	}

	gettermios();
	lastp = cur = r->ifcall.data;
	for(; cur < r->ifcall.data + r->ifcall.count; cur++) {
		if(*cur == ' ' || cur < r->ifcall.data + r->ifcall.count - 1) {
			if(*cur == ' ')
				*cur = '\0';
			cur++;
			switch(*lastp) {
			case 'b':
				setbaud(lastp + 1);
				break;
			case 'd':
				setdtr(lastp + 1);
				break;
			case 'k':
				setbrk(lastp + 1);
				break;
			case 'l':
				setbpb(lastp + 1);
				break;
			case 'p':
				setparity(lastp + 1);
				break;
			case 'r':
				setrts(lastp + 1);
				break;
			case 's':
				setstopbits(lastp + 1);
				break;
			default:
				break;
			}
			lastp = cur;
		}
	}	
	settermios();

	respond(r, nil);
}

static void
uiastatusread(Req *r)
{
	if(r->ifcall.offset > 0) {
		respond(r, nil);
		return;
	}

	r->ofcall.count = snprint(r->ofcall.data, r->ifcall.count,
			"b%d d%d l%d p%c r%d s%s\n",
			unpack32(termios),
			ctrl & 0x01,
			termios[6] & 0xFF,
			(termios[5] == 1 ? 'o' :
				(termios[5] == 2 ? 'e' :
					(termios[5] == 3 ? 'm' :
						(termios[5] == 4 ? 's' : 'n')
					)
				)
			),
			(ctrl >> 1) & 0x01,
			(termios[4] == 1 ? "1.5" :
				(termios[4] == 2 ? "2" : "1")
			));
	respond(r, nil);
}

static void
fsread(Req *r)
{
	char e[ERRMAX];
	ulong path;
	path = r->fid->qid.path;
	switch(TYPE(path)) {
	case Qroot:
		dirread9p(r, rootgen, nil);
		respond(r, nil);
		break;
	case Quia:
		procrfork(uiaread, r, STACKSIZE, 0);
		break;
	case Quiactl:
		uiactlread(r);
		break;
	case Quiastatus:
		uiastatusread(r);
		break;
	default:
		snprint(e, sizeof(e), "bug in fsread path=%lux", path);
		respond(r, e);
		break;
	}
}

static void
fswrite(Req *r)
{
	ulong path;
	char e[ERRMAX];

	path = r->fid->qid.path;
	switch(TYPE(path)) {
	case Quia:
		procrfork(uiawrite, r, STACKSIZE, 0);
		break;
	case Quiactl:
		uiactlwrite(r);
		break;
	default:
		snprint(e, sizeof(e), "bug in fswrite path=%lux", path);
		respond(r, e);
		break;
	}
}

static void
fsopen(Req *r)
{
	static int need[4] = {4, 2, 6, 1};
	ulong path;
	int n;
	Tab *t;

	path = r->fid->qid.path;
	t = &tab[TYPE(path)];
	n = need[r->ifcall.mode & 3];
	if((n & t->mode) != n) {
		respond(r, "permission denied");
		return;
	}

	switch(TYPE(path)) {
	case Quia:
	case Qroot:
	case Quiactl:
	case Quiastatus:
		respond(r, nil);
		break;
	default:
		respond(r, "bug in fsopen");
		break;
	}
}

static void
fsflush(Req *r)
{
	Req *cur, *spe;

	spe = nil;

	qlock(&relock);
	for(cur = rq; cur != nil; cur = cur->aux) {
		if(cur->ifcall.tag == r->ifcall.oldtag) {
			respond(cur, "flush");
			if(cur == rq)
				rq = cur->aux;
			if(cur == *erq) {
				if(spe != nil) {
					erq = &spe;
					spe->aux = nil;
				} else {
					rq = nil;
					erq = nil;
				}
			}
		}
		spe = cur;
	}
	qunlock(&relock);
	respond(r, nil);
}

Srv fs =
{
.flush =	fsflush,
.attach =	fsattach,
.walk1 =	fswalk1,
.open =		fsopen,
.read =		fsread,
.write =	fswrite,
.stat =		fsstat,
};

void
usage(void)
{
	fprint(2, "usage: %s [-dD] [-m mtpt] [-s srv] [ctlrno usbport]\n", argv0);
	threadexitsall("usage");
}

void
threadmain(int argc, char *argv[])
{
	int ctlrno, i, sfd;
	char line[256], *service, *mtpt;

	ctlrno = 0;
	i = 0;
	mtpt = "/n/uart";
	service = nil;

	ARGBEGIN {
	case 'd':
		debug = 1;
		break;
	case 'D':
		chatty9p++;
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	case 's':
		service = EARGF(usage());
		break;
	default:
		usage();
	} ARGEND;

	switch(argc) {
	case 0:
		for(ctlrno = 0; ctlrno < 16; ctlrno++) {
			for(i = 0; i < 128; i++) {
				snprint(line, sizeof(line), cfmt, ctlrno, i, "status");
				sfd = open(line, OREAD);
				if(sfd < 0)
					break;
				if(read(sfd, line, strlen(ident)) &&
					strncmp(ident, line, strlen(ident)) == 0) {
					if(debug)
						fprint(2, "Found matching device at /dev/usb%d/%d\n", ctlrno, i);
					close(sfd);
					goto found;
				}
				close(sfd);
			}
		}
		if(debug)
			fprint(2, "No matching device found.\n");
		threadexitsall("no device found");
	found:
		break;
	case 2:
		ctlrno = atoi(argv[0]);
		i = atoi(argv[1]);
		break;
	default:
		usage();
	}

	snprint(line, sizeof(line), cfmt, ctlrno, i, "ctl");
	sfd = open(line, OWRITE);
	if(sfd < 0)
		sysfatal("%s: %r", line);
	if(debug)
		fprint(2, "Send %s to %s\n", instr, line);
	write(sfd, instr, strlen(instr));
	if(debug)
		fprint(2, "Send %s to %s\n", outstr, line);
	write(sfd, outstr, strlen(outstr));
	close(sfd);

	snprint(line, sizeof(line), cfmt, ctlrno, i, "setup");
	statfd = open(line, ORDWR);
	if(statfd < 0)
		sysfatal("%s: %r", line);
	if(debug)
		fprint(2, "Start reading and writing from %s\n", line);

	snprint(line, sizeof(line), cfmt, ctlrno, i, "ep2data");
	outfd = open(line, OWRITE);
	if(outfd < 0)
		sysfatal("%s: %r", line);
	if(debug)
		fprint(2, "Start writing on %s\n", line);

	snprint(line, sizeof(line), cfmt, ctlrno, i, "ep3data");
	infd = open(line, OREAD);
	if(infd < 0)
		sysfatal("%s: %r", line);
	if(debug)
		fprint(2, "Start reading on %s\n", line);

	initpl2303();
	time0 = time(0);

	proccreate(fsreadproc, nil, STACKSIZE);

	threadpostmountsrv(&fs, service, mtpt, MREPL);

	threadexits(nil);
}
